# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
import itertools
import multiprocessing as mp
import os
import tkinter as tk
from functools import wraps

import numpy as np
from abc import abstractmethod

from fototex import MAX_NB_OF_SAMPLED_FREQUENCIES, NB_PCA_COMPONENTS, R_SPECTRA_NO_DATA_VALUE, \
    _ISOTROPIC_R_SPECTRA_AXIS, _ISOTROPIC_NB_SAMPLE_AXIS, _ANISOTROPIC_R_SPECTRA_AXIS, \
    _ANISOTROPIC_NB_SAMPLE_AXIS, MAX_NB_SECTORS, R_SPECTRA_H5_DS_NAME, \
    REDUCED_R_SPECTRA_H5_DS_NAME, EIGEN_VECTOR_H5_DS_NAME, WINDOW_SIZE_ATTR_NAME, METHOD_ATTR_NAME, LOADINGS_H5_DS_NAME, \
    EXPLAINED_VARIANCE_RATIO_DS_NAME
from fototex._numba import get_sector_directions, get_valid_values
from fototex._tools import mp_r_spectra, mp_h5_r_spectra, normal_pca, \
    h5_incremental_pca, h5_pca_transform, h5_incremental_pca_sector, \
    normal_pca_sector, mp_h5_r_spectra_sector, mp_r_spectra_sector, pca_transform
from fototex.exceptions import FotoBaseError
from fototex.io import H5File, H5TempFile, write_h5
from fototex.plotting import plot_correlation_circle
from fototex.utils import lazyproperty, check_string, check_type, isdir


# TODO: save all FOTO data (eigen vectors, r-spectra and reduced r-spectra) in one unique HDF5 file


def boolean(setter):

    @wraps(setter)
    def _bool(self, value):
        try:
            check_type(value, bool)
        except TypeError:
            raise FotoBaseError("'%s' must be boolean but is: '%s'" %
                                (setter.__name__, type(value).__name__))
        output = setter(self, value)

    return _bool


def directory(setter):

    @wraps(setter)
    def _directory(self, path):
        if not isdir(path):
            raise FotoBaseError("'%s' must be a valid directory path" % setter.__name__)
        output = setter(self, path)

    return _directory


def integer(setter):

    @wraps(setter)
    def _integer(self, value):
        try:
            check_type(value, int)
        except TypeError:
            raise FotoBaseError("'%s' must be an integer value but is: '%s'" %
                                (setter.__name__, type(value).__name__))
        output = setter(self, value)

    return _integer


def odd(setter):

    @wraps(setter)
    def _odd(self, value):
        if value % 2 == 0:
            raise FotoBaseError("'%s' must be an odd value (=%d)" % (setter.__name__, value))
        output = setter(self, value)

    return _odd


def positive(setter):

    @wraps(setter)
    def _positive(self, value):
        if value <= 0:
            raise FotoBaseError("'%s' must be positive (=%d)" % (setter.__name__, value))
        output = setter(self, value)

    return _positive


class FotoBase:
    """ Foto base class

    Main Foto abstract class for all Foto subclasses
    """
    _r_spectra = None
    _r_spectra_reduced = None
    _in_memory = None
    _keep_dc_component = None
    _method = None
    _out_dir = None
    _nb_sampled_frequencies = None
    _normalized = None
    _standardized = None
    _window_size = None
    _window_step = None

    data_chunk_size = None
    max_nb_sampled_frequencies = MAX_NB_OF_SAMPLED_FREQUENCIES
    nb_pca_components = NB_PCA_COMPONENTS
    no_data_value = R_SPECTRA_NO_DATA_VALUE
    eigen_vectors = None
    loadings = None
    explained_variance_ratio = None

    r_spectra_axis = _ISOTROPIC_R_SPECTRA_AXIS
    nb_sample_axis = _ISOTROPIC_NB_SAMPLE_AXIS

    def _compute_pca(self, *args, **kwargs):
        if self.in_memory:
            self.eigen_vectors, \
                self.loadings, \
                self._r_spectra_reduced, \
                self.explained_variance_ratio = normal_pca(self)
        else:
            self.eigen_vectors, \
                self.loadings, \
                self.explained_variance_ratio = h5_incremental_pca(self, *args, **kwargs)

    def _compute_r_spectra(self, nb_processes, *args, **kwargs):
        if self.in_memory:
            self._r_spectra = mp_r_spectra(self, nb_processes)
        else:
            mp_h5_r_spectra(self, nb_processes)

    @abstractmethod
    def _plot_factorial_plan(self, root, window_size, reduced_r_spectra,
                             method, nb_points, data_range, nb_quadrants,
                             norm_method, nb_windows_per_side,
                             main_fig_rel_size, contrast_range, invert_axis,
                             *args, **kwargs):
        pass

    def compute_pca(self, standardized=True, at_random=False,
                    batch_size=None, max_iter=1000, *args, **kwargs):
        """ Compute PCA for r-spectra tables

        Description
        -----------
        Reduce dimensionality of r-spectra table, with
        respect to number of components (by default set
        to 3 for RGB maps calculation), by applying
        principal component analysis (PCA)

        Parameters
        ----------
        standardized : bool
            if True, standardize r-spectrum data before PCA (subtract mean and divide by std)
        at_random: bool
            apply random incremental pca
        batch_size: int
            size of batch for random incremental pca if at_random=True
        max_iter: int
            maximum number of iterations if at_random=True

        Returns
        -------
        FotoBase:
            the current instance
        """
        self.standardized = standardized
        self._compute_pca(at_random, batch_size, max_iter)

    def compute_r_spectra(self, window_size, window_step=1, nb_sampled_frequencies=None,
                          normalized=False, keep_dc_component=False, nb_processes=mp.cpu_count(),
                          *args, **kwargs):
        """ Compute r-spectra over image with respect to sliding window method

        Description
        -----------
        Compute rspectra tables for given image,
        depending on the selected sliding window
        method and other parameters (window size,
        standardization, etc.)

        Parameters
        ---------
        window_size: int
            size of window
        window_step: int, default 1
            step used in sliding window if method is "moving_window"
        nb_sampled_frequencies: int
            number of sampled frequencies (optional)
            If None, is inferred from window size
        normalized: bool
            if True, divide by window variance
        keep_dc_component: bool
            keep the DC component (0 frequency) of the FFT. Use carefully as it may
            substantially change the final results!
        nb_processes: int
            number of processes for multiprocessing calculation

        Returns
        -------
        FotoBase:
            the current instance with r-spectra that have been computed
        """
        # Set window size and standardize bool
        self.window_size = window_size
        self.window_step = window_step
        self.normalized = normalized
        self.keep_dc_component = keep_dc_component

        if not nb_sampled_frequencies:
            self.nb_sampled_frequencies = min(int(window_size / 2),
                                              self.max_nb_sampled_frequencies)
        else:
            self.nb_sampled_frequencies = nb_sampled_frequencies

        self._compute_r_spectra(nb_processes)

    def fit_transform(self, other, nb_processes=mp.cpu_count(), *args, **kwargs):
        """ Apply eigen vectors from other Foto object to current object's R-spectra

        Description
        -----------
        Use the PCA eigen vectors retrieved from
        another FotoBase class that has been run
        in order to get reduced r-spectra for the
        current instance

        Examples
        --------

        Parameters
        ----------
        other : FotoBase
            FotoBase class instance that have been run
        nb_processes : int
            number of processes to open for multiprocessing

        Returns
        -------
        FotoBase:
            the current instance
        """
        # TODO: prototype method (apply eigen vectors computed from some image to another)

        # Compute r-spectra and project in input Foto eigenvector's base
        self.compute_r_spectra(other.window_size,
                               nb_sampled_frequencies=other.nb_sampled_frequencies,
                               normalized=other.normalized,
                               nb_processes=nb_processes)

        if self.in_memory:
            self._r_spectra_reduced = pca_transform(self, other)
        else:
            h5_pca_transform(self, other)

    @abstractmethod
    def get_window_generator(self):
        pass

    def plot_correlation_circle(self, h5path=None, unit='km', invert_axis=False, fontsize=12, circle=True):
        """ Plot correlation circle

        Parameters
        ----------
        h5path: str
            Path to HDF5 file where reduced r-spectra
            are stored. If None, instance r-spectra are
            used. Useful for later plotting.
        unit: str
            'n_occ' or 'km'
        invert_axis: bool
            if True, set PCA 1 on y axis / PCA 2 on x axis
        fontsize: int
            Text size of frequency labels
        circle: bool
            If True, plot circle

        Returns
        -------

        """
        if h5path is not None:
            h5file = H5File(h5path)
            loadings = h5file[LOADINGS_H5_DS_NAME]
        else:
            loadings = self.loadings

        plot_correlation_circle(self, loadings, unit, invert_axis, fontsize, circle)

    def plot_factorial_plan(self, h5path=None, nb_points=10000, data_range=None,
                            nb_quadrants=12, norm_method="max", nb_windows_per_side=2,
                            main_fig_rel_size=0.6, contrast_range=(2, 98), invert_axis=False,
                            *args, **kwargs):
        """ Plot factorial plan and corresponding windows for each quadrant

        Parameters
        ----------
        h5path: str
            Path to HDF5 file where reduced r-spectra
            are stored. If None, instance r-spectra are
            used. Useful for later plotting.
        nb_points: int
            Number of PC points to be plotted
        data_range: list[float, float]
            Data range as cumulative cut count for
            PC axis1 and axis2 (e.g. [2, 98])
        nb_quadrants: int
            Number of quadrants the factorial plan must
            be divided in
        norm_method: str
            Method used to retrieve windows in each quadrant:
            'max' retrieve window(s) with respect to maximum norm
            'random' retrieve window(s) at random in quadrant
        nb_windows_per_side: int
            Number of windows per side (Each quadrant corresponds
            to a square set of windows such as 1x1, 2x2, 3x3, etc.)
        main_fig_rel_size: float
            Relative size of central figure between 0 and 1
        contrast_range: list[float, float]
            Percentile contrast range used to render
            windows with respect to the whole image.
            ex.: enhance contrast based on cumulative
            count cut between 2% and 98% --> [2, 98]
        invert_axis: bool
            if True, set PCA 1 on y axis / PCA 2 on x axis

        Returns
        -------

        """
        # Initialize tkinter window
        root = tk.Tk()

        # Get reduced r-spectra from current instance or from
        # formerly stored HDF5 dataset
        if h5path is not None:
            h5file = H5File(h5path)
            reduced_r_spectra = h5file[REDUCED_R_SPECTRA_H5_DS_NAME]
            window_size = h5file[REDUCED_R_SPECTRA_H5_DS_NAME].attrs[WINDOW_SIZE_ATTR_NAME]
            method = h5file[REDUCED_R_SPECTRA_H5_DS_NAME].attrs[METHOD_ATTR_NAME]
        else:
            reduced_r_spectra = self.r_spectra_reduced
            window_size = self.window_size
            method = self.method

        # Call plot protected method
        self._plot_factorial_plan(root, window_size, reduced_r_spectra, method, nb_points, data_range,
                                  nb_quadrants, norm_method, nb_windows_per_side, main_fig_rel_size,
                                  contrast_range, invert_axis, *args, **kwargs)

        tk.mainloop()

    def run(self, window_size, window_step=1, nb_sampled_frequencies=None, standardized=True,
            normalized=False, keep_dc_component=False, at_random=False, batch_size=None,
            max_iter=1000, nb_processes=mp.cpu_count(), *args, **kwargs):
        """ Run FOTO algorithm

        Description
        -----------
        Run the whole FOTO algorithm, consisting
        in computing r-spectra and applying PCA,
        with respect to window size and corresponding
        method ("block" or "moving window")

        Parameters
        ----------
        window_size : int
            size of the window for 2-D FFT (must be an odd number when method = "moving")
        window_step: int
            Step used in sliding window if method is "moving_window"
        nb_sampled_frequencies : int
            number of sampled frequencies (if None, is inferred)
        standardized : bool
            if True, standardize r-spectrum data before PCA
        normalized : bool
            if True, divide power spectrum density by window's variance
        keep_dc_component : bool
            either keep or not the DC component (0 frequency) part of the signal FFT. Use it
            carefully as it may change substantially the final results !
        at_random: bool
            if True, use random incremental pca
        batch_size: int
            size of batch for random incremental pca (if None, batch size is inferred)
        max_iter: int
            maximum number of iterations when using random incremental PCA
        nb_processes: int
            number of processes for parallelization

        Examples
        --------
        >>> FotoBase.run(window_size=11, keep_dc_component=True)
        >>> FotoBase.run(window_size=15, standardized=False, normalized=True)

        Returns
        -------
        FotoBase :
            the current instance
        """
        self.compute_r_spectra(window_size,
                               window_step,
                               nb_sampled_frequencies,
                               normalized,
                               keep_dc_component,
                               nb_processes)
        self.compute_pca(standardized, at_random, batch_size, max_iter)

    def save_data(self):
        """ Save FOTO data to H5 file

        Returns
        -------

        """
        self.save_eigen_vectors()
        self.save_r_spectra()
        self.save_reduced_r_spectra()

    def save_eigen_vectors(self):
        """ Save eigen vectors computed with PCA

        Description
        -----------
        Write eigen vectors retrieved from PCA
        to H5 file

        Returns
        -------
        """
        write_h5(self,
                 self.eigen_vectors,
                 EIGEN_VECTOR_H5_DS_NAME)

    def save_explained_variance(self):
        """ Save explained variance ratios computed from PCA

        Returns
        -------

        """
        write_h5(self,
                 self.explained_variance_ratio,
                 EXPLAINED_VARIANCE_RATIO_DS_NAME)

    def save_loadings(self):
        """ Save loadings computed with PCA

        Description
        -----------

        Returns
        -------

        """
        write_h5(self,
                 self.loadings,
                 LOADINGS_H5_DS_NAME)

    def save_r_spectra(self):
        """ Save r-spectra table to h5 file

        Description
        -----------
        Write computed r-spectra to H5 file

        Returns
        -------
        """
        write_h5(self,
                 self._r_spectra,
                 R_SPECTRA_H5_DS_NAME,
                 [WINDOW_SIZE_ATTR_NAME, METHOD_ATTR_NAME],
                 [self.window_size, self.method])

    def save_reduced_r_spectra(self):
        """ Save reduced r-spectra to h5 file

        Returns
        -------

        """
        write_h5(self,
                 self._r_spectra_reduced,
                 REDUCED_R_SPECTRA_H5_DS_NAME,
                 [WINDOW_SIZE_ATTR_NAME, METHOD_ATTR_NAME],
                 [self.window_size, self.method])

    @property
    def chunk_size(self):
        return int(self.data_chunk_size / self.nb_sampled_frequencies)

    @property
    def cycles_km(self):
        """ Return sampled frequencies in cycles.km-1

        Returns
        -------

        """
        return [np.round((f * 1000) / (self.window_size * self.pixel_size))
                for f in range(1, self.nb_sampled_frequencies + 1)]

    @lazyproperty
    def h5(self):
        if not self.in_memory:
            return H5TempFile()

    @lazyproperty
    def h5_file(self):
        return H5File(self.path + "foto_data.h5")

    @property
    def in_memory(self):
        return self._in_memory

    @in_memory.setter
    @boolean
    def in_memory(self, value):
        self._in_memory = value

    @property
    def keep_dc_component(self):
        return self._keep_dc_component

    @keep_dc_component.setter
    @boolean
    def keep_dc_component(self, value):
        self._keep_dc_component = value

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, value):
        try:
            self._method = check_string(value, {'block', 'moving_window'})
        except (TypeError, ValueError) as e:
            raise FotoBaseError("Invalid sliding window method: '%s'" % value)

    @property
    @abstractmethod
    def nb_windows(self):
        pass

    @property
    @abstractmethod
    def gdal_no_data_value(self):
        pass

    @property
    @abstractmethod
    def pixel_size(self):
        pass

    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    @directory
    def out_dir(self, path):
        self._out_dir = path

    @property
    def path(self):
        return os.path.join(self.out_dir, f"{self.__class__.__name__.upper()}_method"
                                          f"={self.method}_wsize="
                                          f"{self.window_size}_wstep={self.window_step}_"
                                          f"dc={self.keep_dc_component}_normalized={self.normalized}_")

    @property
    def mean(self):
        if self.in_memory:
            return get_valid_values(self._r_spectra, self.no_data_value).mean(axis=0)
        else:
            return self.h5[R_SPECTRA_H5_DS_NAME].attrs['mean']

    @property
    def nb_sampled_frequencies(self):
        return self._nb_sampled_frequencies

    @nb_sampled_frequencies.setter
    @integer
    @positive
    def nb_sampled_frequencies(self, value):
        if value < self.nb_pca_components:
            raise ValueError("You cannot sample less frequencies than PCA components ! "
                             "(you may want to check window size)")
        self._nb_sampled_frequencies = value

    @property
    def normalized(self):
        return self._normalized

    @normalized.setter
    @boolean
    def normalized(self, value):
        self._normalized = value

    @property
    def std(self):
        if self.in_memory:
            return get_valid_values(self._r_spectra, self.no_data_value).std(axis=0)
        else:
            return self.h5[R_SPECTRA_H5_DS_NAME].attrs['std']

    @property
    def r_spectra(self):
        if self.in_memory:
            return self._r_spectra
        else:
            return self.h5[R_SPECTRA_H5_DS_NAME]

    @property
    def r_spectra_reduced(self):
        if self.in_memory:
            return self._r_spectra_reduced
        else:
            return self.h5[REDUCED_R_SPECTRA_H5_DS_NAME]

    @property
    def standardized(self):
        return self._standardized

    @standardized.setter
    @boolean
    def standardized(self, value):
        self._standardized = value

    @property
    def window_size(self):
        return self._window_size

    @window_size.setter
    @integer
    @odd
    @positive
    def window_size(self, value):
        self._window_size = value

    @property
    def window_step(self):
        return self._window_step

    @window_step.setter
    @integer
    @positive
    def window_step(self, value):
        self._window_step = value


class Batch(FotoBase):
    """ Batch abstract class

    Description
    -----------
    Batch is used to allow the FOTO algorithm
    to be applied from multiple image batches.
    This class should not be used, but be
    inherited by subclasses that must implement
    the batch process, especially by using
    multiple inheritance
    """

    foto_instances = None

    def get_window_generator(self):
        w_gen = []
        for foto in self.foto_instances:
            w_gen.append(foto.get_window_generator())

        return itertools.chain(*w_gen)

    @property
    @abstractmethod
    def gdal_no_data_value(self):
        pass

    @property
    def nb_windows(self):
        nb_rspec = 0
        for foto in self.foto_instances:
            nb_rspec += foto.nb_windows
        return nb_rspec


class Sector(FotoBase):
    """ Sector abstract method to implement anisotropy within FOTO algorithm

    Note
    ----
    In order to correctly override FotoBase and use Sector with
    other subclasses of FotoBase, we must keep the constructor
    syntax order. That is to put "nb_sectors" at the end. In case
    we later use multiple inheritance, it is very important so that
    there is no conflict between multiple constructors inheriting
    from the same superclass (here FotoBase)
    """

    nb_sectors = None
    start_sector = None
    r_spectra_axis = _ANISOTROPIC_R_SPECTRA_AXIS
    nb_sample_axis = _ANISOTROPIC_NB_SAMPLE_AXIS
    max_nb_sectors = MAX_NB_SECTORS

    def _compute_pca(self, *args, **kwargs):
        if self.in_memory:
            self.eigen_vectors, self._r_spectra_reduced = normal_pca_sector(self)
        else:
            self.eigen_vectors = h5_incremental_pca_sector(self, *args, **kwargs)

    def _compute_r_spectra(self, nb_processes, *args, **kwargs):
        if self.in_memory:
            self._r_spectra = mp_r_spectra_sector(self, nb_processes)
        else:
            self._r_spectra = mp_h5_r_spectra_sector(self, nb_processes)

    def save_eigen_vectors(self):
        """ Save eigen vectors from PCA decomposition

        Description
        -----------

        Returns
        -------
        :return:
        """
        for path, eigen_vectors in zip(self.path, self.eigen_vectors):
            np.savetxt(path + "eigen_vectors.csv", eigen_vectors, delimiter=',')

    @property
    def chunk_size(self):
        return int(self.data_chunk_size / (self.nb_sampled_frequencies * self.nb_sectors))

    @lazyproperty
    def sectors(self):
        return get_sector_directions(self.nb_sectors, self.start_sector) * 180 / np.pi

    @abstractmethod
    def get_window_generator(self):
        pass

    @property
    @abstractmethod
    def nb_windows(self):
        pass



# -*- coding: utf-8 -*-

""" Module gathering all numba-based optimized functions

We preferentially use numba jit decorator without signature. As detailed
in the numba documentation in that case:
"The decorated function implements lazy compilation.
Each call to the decorated function will try to re-use an existing
specialization if it exists (for example, a call with two integer
arguments may re-use a specialization for argument types
(numba.int64, numba.int64)). If no suitable specialization exists,
a new specialization is compiled on-the-fly, stored for later use,
and executed with the converted arguments."
"""

import numpy as np

from numba import jit, int64


@jit(nopython=True, nogil=True)
def contains(array, value):
    for n in array.ravel():
        if n == value:
            return True
    return False


@jit(nopython=True, nogil=True)
def only_contains(array, value):
    for n in array.ravel():
        if n != value:
            return False
    return True


# @jit([(float64[:, :], ), (float32[:, :], )], nopython=True, nogil=True)
@jit(nopython=True, nogil=True)
def upper_neighbor(array):
    """ Replace NaNs in given 2D array column by upper neighbor

    Replace NaNs in place in 2D array by
    using nearest upper neighbor
    :param array:
    :return:
    """
    # Using a loop with numba is faster
    for n, value in enumerate(array.ravel()):
        if np.isnan(value):
            array.ravel()[n] = array.ravel()[n - array.shape[1]]

    # Numpy-based method (slower)
    # no_value = np.where(np.isnan(array.ravel()))[0]
    # upper = no_value - array.shape[1]
    # array.ravel()[no_value] = array.ravel()[upper]


# @jit([float64[:](int64[:], float64[:]),
# float32[:](int64[:], float32[:])], nopython=True, nogil=True)
@jit(nopython=True, nogil=True)
def get_average_per_radius(radius, values):
    """ Get average of all values with the same radius

    :param radius:
    :param values:
    :return:
    """
    average = np.bincount(radius, values) / np.bincount(radius)

    return average.astype(values.dtype)


# @jit([(float64[:, :], int64), (float32[:, :], int64)], nopython=True, nogil=True)
@jit(nopython=True, nogil=True)
def get_no_nan(array, idx_col):
    """ Get first no NaN value in column

    :param array:
    :param idx_col:
    :return:
    """
    for value in array[:, idx_col]:
        if ~np.isnan(value):
            return value


# @jit([(float64[:, :], int64[:], int64[:], int64),
#       (float32[:, :], int64[:], int64[:], int64)], nopython=True, nogil=True)
@jit(nopython=True, nogil=True)
def sector_average(window, radius, sectors, nb_sectors):
    window = window.ravel()
    r_average_per_sector = np.zeros((nb_sectors, radius.max() + 1))
    # for sector in np.unique(sectors):
    for sector in range(1, nb_sectors + 1):
        mask = np.where(sectors == sector)
        average = get_average_per_radius(radius[mask], window[mask])
        r_average_per_sector[sector - 1, 0:average.size] = average

    # r_average_per_sector[:, 0] = r_average_per_sector[~np.isnan(r_average_per_sector[:, 0]), 0]
    r_average_per_sector[:, 0] = get_no_nan(r_average_per_sector, 0)
    upper_neighbor(r_average_per_sector)

    return r_average_per_sector


# @jit([float64[:](int64[:], float64[:, :]), float32[:](int64[:],
# float32[:, :])], nopython=True, nogil=True)
@jit(nopython=True, nogil=True)
def azimuthal_average(radius, window):

    return get_average_per_radius(radius, window.ravel())


@jit(nopython=True, nogil=True)
def azimuthal_max(radius, window):

    return get_max_per_radius(radius, window)


@jit(nopython=True, nogil=True)
def get_max_per_radius(radius, window):
    """ Get maximum value in window over each radius

    :param radius:
    :param window:
    :return:
    """
    len_radius = radius.max()
    all_max = np.zeros(len_radius)
    window_view = window.ravel()
    for rad in range(len_radius):
        all_max[rad] = window_view[radius == rad].max()

    return all_max


# @jit(float64[:, :](int64[:, :], int64[:, :]), nopython=True, nogil=True)
@jit(nopython=True, nogil=True)
def get_azimuth(x, y):
    """ Return azimuth of all window pixels with respect to window origin

    :param x: pixel x coordinates as returned by np.indices function
    :param y: pixel y coordinates as returned by np.indices function
    :return:
    """
    return np.arctan2(x - (x.shape[1] - 1) / 2, (y.shape[0] - 1) / 2 - y)


# @jit([float64[:](int64, int64), float64[:](int64, float64)], nopython=True, nogil=True)
@jit(nopython=True, nogil=True)
def get_bin_sectors(nb_bins, center):
    """ Return bin circle sectors from required number of bins

    Return bin sectors centered on specified direction (default North, i.e 0°)
    :param nb_bins: (integer) number of bins
    :param center: center of the bin sectors in radians
    :return:
    """
    bin_vector = np.linspace(-(nb_bins - 1) * np.pi / nb_bins,
                             (nb_bins - 1) * np.pi / nb_bins, nb_bins)
    bin_quadrants = bin_vector + (nb_bins % 2) * np.pi/nb_bins + (center % (2*np.pi))
    bin_quadrants[bin_quadrants < np.pi] = bin_quadrants[bin_quadrants < np.pi] + 2*np.pi
    bin_quadrants[bin_quadrants > np.pi] = bin_quadrants[bin_quadrants > np.pi] - 2*np.pi
    bin_quadrants.sort()

    return bin_quadrants


# @jit([float64[:](int64, int64), float64[:](int64, float64)], nopython=True, nogil=True)
@jit(nopython=True, nogil=True)
def get_sector_directions(nb_sectors, start):
    """ Get direction of each sector

    Index of direction in array gives the bin number
    in get_sectors function (starting from 1)
    :param nb_sectors: number of sectors
    :param start: starting direction
    :return:
    """
    bins = get_bin_sectors(nb_sectors, start)
    inter_bins = (bins[0:-1] + bins[1::]) / 2
    inter_bins = np.asarray(list(inter_bins) + [inter_bins[-1] + 2 * np.pi / nb_sectors])
    inter_bins[inter_bins < 0] = inter_bins[inter_bins < 0] + 2 * np.pi

    return inter_bins


@jit(nopython=True, nogil=True)
def get_block_windows(window_size, raster_x_size, raster_y_size):
    """ Get block window coordinates

    Get block window coordinates depending
    on raster size and window size

    Parameters
    ----------
    window_size: int
        window size
    raster_x_size: int
        raster width
    raster_y_size: int
        raster height

    Returns
    -------

    """
    if raster_x_size % window_size:
        raster_x_size -= (raster_x_size % window_size)
    if raster_y_size % window_size:
        raster_y_size -= (raster_y_size % window_size)

    for y in range(0, raster_y_size, window_size):
        # ysize = min(window_size, raster_y_size - y)
        for x in range(0, raster_x_size, window_size):
            # xsize = min(window_size, raster_x_size - x)

            # yield x, y, xsize, ysize
            yield x, y, window_size, window_size


@jit(nopython=True, nogil=True)
def get_moving_windows(window_size, raster_x_size, raster_y_size, step=1):
    """ Get moving window coordinates

    Get moving window coordinates depending
    on raster size, window size and step

    Parameters
    ----------
    window_size: int
    raster_x_size: int
    raster_y_size: int
    step: int

    Returns
    -------

    """
    offset = int((window_size - 1) / 2)  # window_size must be an odd number
    # for each pixel, compute indices of the window (all included)

    if raster_x_size % step:
        raster_x_size -= (raster_x_size % step)
    if raster_y_size % step:
        raster_y_size -= (raster_y_size % step)

    for y in range(offset, raster_y_size - offset, step):
        y1 = y - offset
        y2 = y + offset
        ysize = (y2 - y1) + 1
        for x in range(offset, raster_x_size - offset, step):
            x1 = x - offset
            x2 = x + offset
            xsize = (x2 - x1) + 1

            yield x1, y1, xsize, ysize


# @jit([int64[:](int64[:, :], int64[:, :], int64, int64),
# int64[:](int64[:, :], int64[:, :], int64, float64)],
#      nopython=True, nogil=True)
@jit(nopython=True, nogil=True)
def get_sectors(x, y, nb_sectors, center):
    sectors = np.digitize(get_azimuth(x, y), get_bin_sectors(nb_sectors, center))
    sectors = sectors.ravel()
    mask = np.where(sectors == 0)
    sectors[mask] = nb_sectors

    return sectors


# @jit(int64[:](int64[:, :], int64[:, :]), nopython=True, nogil=True)
@jit(nopython=True, nogil=True)
def get_radius(x, y):
    """ Return radius of window pixels with respect to window origin

    :param x: pixel x coordinates
    :param y: pixel y coordinates
    :return:
    """
    radius = np.sqrt((x - (x.shape[1] - 1) / 2) ** 2 +
                     (y - (x.shape[0] - 1) / 2) ** 2).astype(int64)

    return radius.ravel()


# @jit([float64[:, :](float64[:, :], int64),
# float32[:, :](float32[:, :], int64)], nopython=True, nogil=True)
@jit(nopython=True, nogil=True)
def get_valid_values(array, no_data_value):
    """ Get valid values from 2d array

    It is assumed that no data are rows (i.e. axis=0)
    Whenever possible (when there is no "no data")
    return a reference to array
    :param array:
    :param no_data_value: value filling entire rows of the given array
    :return:
    """
    view = array.ravel()
    mask = np.where(view != no_data_value)

    if mask[0].size == view.size:
        return array
    else:
        return view[mask].reshape((mask[0].size // array.shape[1], array.shape[1]))


# @jit([float64[:, :, :](float64[:, :, :], int64), float32[:, :, :](float32[:, :, :], int64)],
#      nopython=True, nogil=True)
@jit(nopython=True, nogil=True)
def get_valid_values_3d(array, no_data_value):
    """ Get valid values from 3D array

    No data are plans normal to axis=1, i.e.
    with width along axis=0 and height along axis=2
    :param array:
    :param no_data_value:
    :return:
    """
    view = array.ravel()
    mask = np.where(view != no_data_value)

    if mask[0].size == view.size:
        return array
    else:
        return view[mask].reshape((array.shape[0],
                                   mask[0].size // (array.shape[0] * array.shape[2]),
                                   array.shape[2]))


@jit(nopython=True, nogil=True)
def zero_padding(window, no_data_value):
    """ Replace window no data by window mean

    Description
    -----------
    Replace window no data values by the mean
    over valid window values (corresponds to 0-padding
    with 0s as the mean over standardized windows)
    """
    new_window = window.copy()
    view = new_window.ravel()
    view[view == no_data_value] = view[view != no_data_value].mean()

    return new_window



# -*- coding: utf-8 -*-

""" Plotting functions for FOTO outputs

"""
import tkinter

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends._backend_tk import NavigationToolbar2Tk

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure

from fototex import R_SPECTRA_NO_DATA_VALUE
from fototex._numba import get_bin_sectors, get_sector_directions
from fototex.foto_tools import azimuth_to_counterclockwise, get_window_over_multiple_images

GRID_X = 1
GRID_Y = 1
PAD_FACTOR = 6
PC_AXIS_1 = 0
PC_AXIS_2 = 1
SUBFIG_PAD = 0.06
TK_WINDOW_SIZE = 1000
DPI = 100
NORTH_DIRECTION = 0
NORM_FONT_SIZE = 8
NORM_FONT_COLOR = "red"


def factorial_plan_plot(root, directions,
                        tk_window_size, fig_rel_size,
                        dpi):

    subfig_rel_height = min((1 - fig_rel_size) / 2 - SUBFIG_PAD,
                            1/(1 + len(directions)/4) - SUBFIG_PAD)
    subfig_rel_width = min((1 - fig_rel_size) / 2 - SUBFIG_PAD,
                           1/(1 + len(directions)/4) - SUBFIG_PAD)

    root.wm_title("Factorial plan")
    root.geometry("%dx%d" % (tk_window_size, tk_window_size))

    pad_x = PAD_FACTOR * fig_rel_size
    pad_y = PAD_FACTOR * fig_rel_size
    grid_x_min = (1 - fig_rel_size) / pad_x - GRID_X / 2
    grid_y_min = (1 - fig_rel_size) / pad_y - GRID_Y / 2
    grid_x_max = (pad_x - 1 + fig_rel_size) / pad_x - GRID_X / 2
    grid_y_max = (pad_y - 1 + fig_rel_size) / pad_y - GRID_Y / 2

    main_fig = Figure(dpi=dpi)
    canvas = FigureCanvasTkAgg(main_fig, master=root)
    canvas.get_tk_widget().place(relx=(1 - fig_rel_size) / 2, rely=(1 - fig_rel_size) / 2,
                                 relheight=fig_rel_size, relwidth=fig_rel_size)

    toolbar = NavigationToolbar2Tk(canvas, root, pack_toolbar=False)
    toolbar.update()

    subfig = []
    canvas_subfig = []

    for i, d in enumerate(directions):

        subfig.append(Figure(dpi=dpi))
        canvas_subfig.append(FigureCanvasTkAgg(subfig[i], master=root))

        if d > 7 * np.pi / 4 or d <= np.pi / 4:
            y = grid_y_max
            x = y / np.tan(azimuth_to_counterclockwise(d))
        elif 3 * np.pi / 4 >= d > np.pi / 4:
            x = grid_x_max
            y = x * np.tan(azimuth_to_counterclockwise(d))
        elif 5 * np.pi / 4 >= d > 3 * np.pi / 4:
            y = grid_y_min
            x = y / np.tan(azimuth_to_counterclockwise(d))
        else:
            x = grid_x_min
            y = x * np.tan(azimuth_to_counterclockwise(d))

        canvas_subfig[i].get_tk_widget().place(relx=x + GRID_X / 2,
                                               rely=GRID_Y / 2 - y,
                                               relwidth=subfig_rel_width,
                                               relheight=subfig_rel_height,
                                               anchor="center")

    toolbar.pack(side="bottom", fill="x")

    return main_fig, subfig, canvas, canvas_subfig


def factorial_plan_quadrants(x, y, nb_quadrants):
    """ Retrieve to which quadrant belong factorial plan points

    Parameters
    ----------
    x: numpy.ndarray
        Axis 2 (horizontal) components
    y: numpy.ndarray
        Axis 1 (vertical) components
    nb_quadrants: int
        Number of quadrants

    Returns
    -------

    """
    sectors = get_bin_sectors(nb_quadrants, 0)
    az = np.arctan2(x, y)
    clusters = np.digitize(az, sectors)
    clusters[clusters == 0] = nb_quadrants

    return clusters


def get_data_range(raster_band, range_min, range_max):
    """ Get data range covered by colormap based on cumulative count cut

    Parameters
    ----------
    raster_band: gdal.Band
        GDAL raster band
    range_min: float
        Minimum relative range value (between 0 and 1)
    range_max: float
        Maximum relative range value (between 0 and 1)

    Returns
    -------
    tuple
        Tuple of data range values as (vmin, vmax)

    """
    min_, max_ = raster_band.ComputeRasterMinMax()
    hist = raster_band.GetHistogram(min_ - 0.5, max_ + 0.5, int(max_ - min_) + 1)
    percentile = np.cumsum(hist) / np.sum(hist)

    vmin = min_ + np.argmax(percentile > range_min)
    vmax = min_ + np.argmax(percentile > range_max) - 1

    return vmin, vmax


def get_norm_indices(axis1, axis2, indices, quadrants,
                     nb_quadrants, nb_values, method):
    """ Retrieve full table indices of PC axis1/axis2 corresponding
    to given norm with respect to method selection

    Parameters
    ----------
    axis1: numpy.ndarray
    axis2: numpy.ndarray
    indices: numpy.ndarray
    quadrants: numpy.ndarray
    nb_quadrants: int
        Number of quadrants
    nb_values: int
        Number of indices to retrieve per quadrant
    method: str
        Method used to retrieve indices corresponding
         to a given norm ('random': at random or
        "max" corresponding to maximum norm values)

    Returns
    -------

    """
    norm = np.sqrt(axis1 ** 2 + axis2 ** 2)

    if method == "random":
        rng = np.random.default_rng()
        norm_idx = [rng.choice(indices[quadrants == n + 1], nb_values, replace=False)
                    for n in range(nb_quadrants)]
    elif method == "max":
        rg_idx = np.arange(0, len(norm))
        norm_idx = [indices[rg_idx[quadrants == n + 1][
            norm[quadrants == n + 1].argsort()[::-1][:nb_values]]]
                    for n in range(nb_quadrants)]
    else:
        norm_idx = []

    return norm, norm_idx


def get_pc_axis(reduced_r_spectra, nb_elements=10000, percentile=None):
    """ Get PC axis values from reduced r-spectra

    Parameters
    ----------
    reduced_r_spectra: h5py._hl.dataset.Dataset or numpy.ndarray
        HDF5 dataset or numpy array of values
    nb_elements: int, default 10000
        Number of elements to select at random in r-spectra table
    percentile: list[float, float]
        If not None, extract cumulative count between percentile bounds

    Returns
    -------
    tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]
    """
    rng = np.random.default_rng()
    indices = np.sort(rng.choice(reduced_r_spectra.shape[0], nb_elements,
                                 replace=False, shuffle=False))
    values = reduced_r_spectra[indices, :]
    indices = indices[values[:, 0] != R_SPECTRA_NO_DATA_VALUE]
    axis1 = values[values[:, 0] != R_SPECTRA_NO_DATA_VALUE, PC_AXIS_1]
    axis2 = values[values[:, 1] != R_SPECTRA_NO_DATA_VALUE, PC_AXIS_2]

    if percentile is not None:
        p = np.percentile(axis1, percentile)
        axis2 = axis2[(axis1 >= p[0]) & (axis1 <= p[1])]
        indices = indices[(axis1 >= p[0]) & (axis1 <= p[1])]
        axis1 = axis1[(axis1 >= p[0]) & (axis1 <= p[1])]

    return axis1, axis2, indices


def plot_correlation_circle(foto, loadings, unit, pca1_on_y_axis, fontsize, circle):
    """ Plot correlation circle

    Thanks to the work of Marc Lang (2018)

    Parameters
    ----------
    foto: fototex.foto.Foto
    loadings: numpy.ndarray
    unit: str
    pca1_on_y_axis: bool
    fontsize: int
    circle: bool

    Returns
    -------

    """
    if unit == 'n_occ':
        frequency_label = range(1, foto.nb_sampled_frequencies + 1)
    elif unit == 'km':
        frequency_label = foto.cycles_km
    else:
        frequency_label = [None for f in range(1, foto.nb_sampled_frequencies + 1)]

    ax = plt.subplots()[1]

    if pca1_on_y_axis:
        pca_x = 1
        pca_y = 0
    else:
        pca_x = 0
        pca_y = 1

    for freq in range(foto.nb_sampled_frequencies):
        x = loadings[freq, pca_x]
        y = loadings[freq, pca_y]
        ax.plot(x, y, '.', color='black')
        if unit:
            ax.text(x, y, f"{frequency_label[freq]}",
                    fontsize=fontsize, color='darkslategrey')

    # Plot circle and legend
    ax.set_xlim(-1, 1)
    ax.set_ylim(-1, 1)

    if circle:
        circle = plt.Circle((0, 0), 1, fill=False)
        ax.add_patch(circle)
        ax.axhline(0, color="gray")
        ax.axvline(0, color="gray")

    ax.set_xlabel(f'PC axis {pca_x + 1}')
    ax.set_ylabel(f'PC axis {pca_y + 1}')
    ax.set_title("PC correlation")
    ax.set_facecolor("ivory")

    plt.show()


def plot_factorial_plan(root, foto, window_size, reduced_r_spectra,
                        sliding_window_method, nb_points, data_range, nb_quadrants,
                        method, main_fig_rel_size, nb_of_windows_per_side, contrast_range,
                        pca1_on_y_axis):
    """ Plot factorial plan and corresponding windows for each quadrant

    Parameters
    ----------
    root: tkinter.Tk
        Tkinter root window
    foto: fototex.foto.Foto or fototex.foto.FotoSector or list
    window_size: int
        Size of window used in computing r-spectra
    reduced_r_spectra: h5py._hl.dataset.Dataset or numpy.ndarray
        HDF5 dataset or numpy array of values
    sliding_window_method: str
        Sliding method used in computing r-spectra ('block' or 'moving')
    nb_points: int
        Number of PC points to be plotted
    data_range: list[float, float]
        Data range as cumulative cut count for
        PC axis1 and axis2 (e.g. [2, 98])
    nb_quadrants: int
        Number of quadrants the factorial plan must
        be divided in
    method: str
        Method used to retrieve windows in each quadrant:
        'max' retrieve window(s) with respect to maximum norm
        'random' retrieve window(s) at random in quadrant
    main_fig_rel_size: float
        Relative size of central figure between 0 and 1
    nb_of_windows_per_side: int
        Number of windows per side (Each quadrant corresponds 
        to a square set of windows such as 1x1, 2x2, 3x3, etc.)
    contrast_range: list[float, float]
        Percentile contrast range used to render
        windows with respect to the whole image.
        ex.: enhance contrast based on cumulative
        count cut between 2% and 98% --> [2, 98]
    pca1_on_y_axis: bool
        If True, pca component 1 is set on y axis while pca
        component 2 is set on x axis. If False, pca component 1
        is set on x axis and pca component 2 is set on y axis.

    Returns
    -------

    """
    directions = get_sector_directions(nb_quadrants, NORTH_DIRECTION)
    factorial_plan = factorial_plan_plot(root, directions, TK_WINDOW_SIZE,
                                         main_fig_rel_size, DPI)
    axis1, axis2, indices = get_pc_axis(reduced_r_spectra,
                                        nb_elements=nb_points,
                                        percentile=data_range)

    pca_x_label = "PC axis 2"
    pca_y_label = "PC axis 1"

    if pca1_on_y_axis is False:
        axis1, axis2 = axis2, axis1
        pca_x_label, pca_y_label = pca_y_label, pca_x_label

    if not isinstance(foto, (list, tuple)):
        foto = [foto]

    vmin = [get_data_range(f.dataset.GetRasterBand(f.band),
                           contrast_range[0]/100,
                           contrast_range[1]/100)[0]
            for f in foto]
    vmax = [get_data_range(f.dataset.GetRasterBand(f.band),
                           contrast_range[0]/100,
                           contrast_range[1]/100)[1]
            for f in foto]

    raster_x_size = [f.dataset.RasterXSize for f in foto]
    raster_y_size = [f.dataset.RasterYSize for f in foto]

    quadrants = factorial_plan_quadrants(axis2, axis1, nb_quadrants)
    norm, norm_idx = get_norm_indices(axis1, axis2, indices, quadrants,
                                      nb_quadrants, nb_of_windows_per_side ** 2,
                                      method)

    main_axes = factorial_plan[0].add_subplot(111)
    main_axes.scatter(axis2, axis1, c=quadrants)
    main_axes.set_ylabel(pca_y_label)
    main_axes.set_xlabel(pca_x_label)

    xabs_max = abs(max(main_axes.get_xlim(), key=abs))
    yabs_max = abs(max(main_axes.get_ylim(), key=abs))
    xy_abs_max = max(xabs_max, yabs_max)
    main_axes.set_xlim(xmin=-xy_abs_max, xmax=xy_abs_max)
    main_axes.set_ylim(ymin=-xy_abs_max, ymax=xy_abs_max)

    for i, idx in enumerate(norm_idx):
        sub_axes = factorial_plan[1][i].add_gridspec(nb_of_windows_per_side,
                                                     nb_of_windows_per_side,
                                                     wspace=0, hspace=0).subplots()
        for n, ax in enumerate(sub_axes.ravel()):
            img_id, window = \
                get_window_over_multiple_images(idx[n],
                                                window_size,
                                                raster_x_size,
                                                raster_y_size,
                                                sliding_window_method)
            image = foto[img_id].dataset.GetRasterBand(
                foto[img_id].band).ReadAsArray(*window)
            ax.imshow(image, cmap="gray", vmin=vmin[img_id],
                      vmax=vmax[img_id], aspect="auto")
            ax.set_title(r"$r=%.1f$" % norm[indices == idx[n]], y=1.0,
                         pad=-8, color=NORM_FONT_COLOR, fontsize=NORM_FONT_SIZE)
            ax.set(xticks=[], yticks=[])
            plt.setp(ax.spines.values(), color="white")
        factorial_plan[1][i].subplots_adjust(0, 0, 1, 1)



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

import multiprocessing as mp
import numpy as np
import time
from functools import partial

from sklearn.decomposition import IncrementalPCA
from tqdm import tqdm

from fototex import R_SPECTRA_PG_DESCRIPTION, R_SPECTRA_SECTOR_PG_DESCRIPTION, \
    INC_PCA_PG_DESCRIPTION, REDUCED_R_SPECTRA_PG_DESCRIPTION, PCA_PG_DESCRIPTION, \
    R_SPECTRA_H5_DS_NAME, REDUCED_R_SPECTRA_H5_DS_NAME, PCA_TRANSFORM_PG_DESCRIPTION, MP_CHUNK_SIZE
from fototex._numba import get_radius, get_sectors, get_valid_values, get_valid_values_3d
from fototex.foto_tools import rspectrum, split_into_chunks, pca, standard_deviation, \
    rspectrum_per_sector, standardize


def incremental_pca(chunks, n_components, no_data_value, axis,
                    population_mean=None, population_std=None,
                    standardized=True, nb_sectors=None):
    """ Incremental PCA

    Description
    -----------

    Parameters
    ----------
    chunks: generator
        iterator over data chunks
    n_components: int
        number of dimensions for PCA
    no_data_value: int
        no data value
    axis: int
        r-spectra axis in data chunks
    population_mean: float
        mean of the entire population from which is extracted each chunk
    population_std: float
        std of the entire population from which is extracted each chunk
    standardized: bool
        True if data must be standardized before pca
    nb_sectors: int
        number of sectors (if None, no sectors)
    Returns
    -------
    """
    if nb_sectors is None:
        ipca = IncrementalPCA(n_components=n_components)
    else:
        ipca = [IncrementalPCA(n_components=n_components)] * nb_sectors

    for chunk in chunks:

        if chunk.ndim == 2:
            chunk = get_valid_values(chunk, no_data_value)
        else:
            chunk = get_valid_values_3d(chunk, no_data_value)

        if standardized:
            if population_mean is None:
                chunk -= np.expand_dims(chunk.mean(axis=axis), axis=axis)
            else:
                chunk -= np.expand_dims(population_mean, axis=axis)

            if population_std is None:
                chunk /= np.expand_dims(chunk.std(axis=axis), axis=axis)
            else:
                chunk /= np.expand_dims(population_std, axis=axis)

        if nb_sectors:
            ipca = [ipca_.partial_fit(sub_chunk) for ipca_, sub_chunk in zip(ipca, chunk)]
        else:
            ipca.partial_fit(chunk)

    return ipca


def pca_transform(foto, other):
    """ Transform, in memory, r-spectra to reduced r-spectra using given PCA eigen vectors

    """
    valid_r_spectra = get_valid_values(foto.r_spectra, foto.no_data_value)
    if other.standardized:
        valid_r_spectra = (valid_r_spectra - other.mean) / other.std
    r_spectra_reduced = np.full((foto.nb_windows, other.nb_pca_components), foto.no_data_value,
                                dtype=foto.r_spectra.dtype)
    r_spectra_reduced[foto.r_spectra[:, 0] != foto.no_data_value, :] = np.dot(valid_r_spectra,
                                                                              other.eigen_vectors)

    return r_spectra_reduced


def h5_pca_transform(foto, other):
    """ Transform R-spectra to reduced R-spectra using given PCA eigen vectors

    """
    nb_chunks = foto.h5[R_SPECTRA_H5_DS_NAME].shape[0] // foto.chunk_size + \
        min(1, foto.h5[R_SPECTRA_H5_DS_NAME].shape[0] % foto.chunk_size)
    foto.h5.reset_dataset(REDUCED_R_SPECTRA_H5_DS_NAME,
                          shape=(foto.h5[R_SPECTRA_H5_DS_NAME].shape[0],
                                 other.eigen_vectors.shape[1]))
    for r_spectra in tqdm(foto.h5.read(R_SPECTRA_H5_DS_NAME, foto.chunk_size),
                          total=nb_chunks, desc=PCA_TRANSFORM_PG_DESCRIPTION):
        r_spectra_reduced = np.full((r_spectra.shape[0], other.nb_pca_components),
                                    foto.no_data_value, dtype=r_spectra.dtype)
        valid_r_spectra = get_valid_values(r_spectra, foto.no_data_value)
        if other.standardized:
            valid_r_spectra = (valid_r_spectra - other.mean) / other.std
        r_spectra_reduced[r_spectra[:, 0] != foto.no_data_value, :] = np.dot(valid_r_spectra,
                                                                             other.eigen_vectors)
        foto.h5.append(REDUCED_R_SPECTRA_H5_DS_NAME, r_spectra_reduced)


# TODO: implement PCA transform for sector R-spectra
def pca_transform_sector():
    pass


def mp_h5_r_spectra(foto, nb_processes):
    """ Parallel processing of r-spectra with hdf5 storage

    """
    nb_sample = foto.nb_sampled_frequencies
    foto.h5.reset_dataset(R_SPECTRA_H5_DS_NAME,
                          (foto.nb_windows, nb_sample))

    # Pixel coordinates within window
    y, x = np.indices((foto.window_size, foto.window_size))
    radius = get_radius(x, y)

    # For statistics computation (population mean and std)
    sum_of_values = np.zeros(nb_sample)
    sum_of_square_values = np.zeros(nb_sample)
    nb_valid_r_spectra = 0

    # Compute r-spectra chunk by chunk
    pg = tqdm(total=foto.nb_windows // foto.chunk_size + 1, desc=R_SPECTRA_PG_DESCRIPTION)
    for chk_nb, window_generator in enumerate(split_into_chunks(foto.get_window_generator(),
                                                                foto.chunk_size)):
        with mp.Pool(processes=nb_processes) as pool:
            r_spectra = np.asarray(list(pool.map(partial(rspectrum, radius=radius,
                                                         window_size=foto.window_size,
                                                         nb_sample=nb_sample,
                                                         normalized=foto.normalized,
                                                         keep_dc_component=foto.keep_dc_component,
                                                         no_data_value=foto.gdal_no_data_value),
                                                 window_generator, chunksize=MP_CHUNK_SIZE)))
        valid_r_spectra = get_valid_values(r_spectra, foto.no_data_value)
        nb_valid_r_spectra += valid_r_spectra.shape[foto.r_spectra_axis]
        sum_of_values += np.sum(valid_r_spectra, axis=foto.r_spectra_axis)
        sum_of_square_values += np.sum(valid_r_spectra**2, axis=foto.r_spectra_axis)
        foto.h5.append(R_SPECTRA_H5_DS_NAME, r_spectra)
        pg.update(1)
    pg.close()

    # Write dataset statistics to attributes
    foto.h5[R_SPECTRA_H5_DS_NAME].attrs['mean'] = sum_of_values / nb_valid_r_spectra
    foto.h5[R_SPECTRA_H5_DS_NAME].attrs['std'] = standard_deviation(nb_valid_r_spectra,
                                                                    sum_of_values,
                                                                    sum_of_square_values)

    # TODO: see if chunksize for Pool.map should not be fixed (500 here)


def mp_h5_r_spectra_sector(foto, nb_processes):
    """ Parallel processing of anisotropic r-spectra with hdf5 storage

    """
    nb_sample = foto.nb_sampled_frequencies

    foto.h5.reset_dataset(R_SPECTRA_H5_DS_NAME,
                          (foto.nb_sectors, foto.nb_windows, nb_sample))
    y, x = np.indices((foto.window_size, foto.window_size))
    radius = get_radius(x, y)
    sectors = get_sectors(x, y, foto.nb_sectors, foto.start_sector)

    # Data used for population statistics
    sum_of_values = np.zeros((foto.nb_sectors, nb_sample))
    sum_of_square_values = np.zeros((foto.nb_sectors, nb_sample))
    nb_valid_r_spectra = 0

    # Compute r-spectra chunk by chunk
    pg = tqdm(total=foto.nb_windows // foto.chunk_size + 1, desc=R_SPECTRA_SECTOR_PG_DESCRIPTION)

    for chk_nb, window_generator in \
            enumerate(split_into_chunks(foto.get_window_generator(), foto.chunk_size)):
        with mp.Pool(processes=nb_processes) as pool:
            r_spectra = list(pool.map(partial(rspectrum_per_sector, radius=radius, sectors=sectors,
                                              window_size=foto.window_size, nb_sample=nb_sample,
                                              nb_sectors=foto.nb_sectors,
                                              normalized=foto.normalized,
                                              keep_dc_component=foto.keep_dc_component,
                                              no_data_value=foto.gdal_no_data_value),
                                      window_generator, chunksize=MP_CHUNK_SIZE))
        r_spectra = np.transpose(np.asarray(r_spectra), (1, 0, 2))
        valid_r_spectra = get_valid_values_3d(r_spectra, foto.no_data_value)
        nb_valid_r_spectra += valid_r_spectra.shape[foto.r_spectra_axis]
        sum_of_values += np.sum(valid_r_spectra, axis=foto.r_spectra_axis)
        sum_of_square_values += np.sum(valid_r_spectra ** 2, axis=foto.r_spectra_axis)
        foto.h5.append(R_SPECTRA_H5_DS_NAME,
                       r_spectra, axis=foto.r_spectra_axis)
        pg.update(1)
    pg.close()

    # Write dataset statistics to attributes
    foto.h5[R_SPECTRA_H5_DS_NAME].attrs['mean'] = sum_of_values / nb_valid_r_spectra
    foto.h5[R_SPECTRA_H5_DS_NAME].attrs['std'] = standard_deviation(nb_valid_r_spectra,
                                                                    sum_of_values,
                                                                    sum_of_square_values)


def mp_r_spectra(foto, nb_processes):
    """ Parallel processing of isotropic r-spectra in memory

    """
    nb_sample = foto.nb_sampled_frequencies

    # Pixel coordinates within window
    y, x = np.indices((foto.window_size, foto.window_size))
    radius = get_radius(x, y)

    # Parallel computation of r-spectra
    with mp.Pool(processes=nb_processes) as pool:
        r_spectra = list(tqdm(pool.imap(partial(rspectrum, radius=radius,
                                                window_size=foto.window_size,
                                                nb_sample=nb_sample, normalized=foto.normalized,
                                                keep_dc_component=foto.keep_dc_component,
                                                no_data_value=foto.gdal_no_data_value),
                                        foto.get_window_generator(), chunksize=MP_CHUNK_SIZE),
                              total=foto.nb_windows, unit_scale=True,
                              desc=R_SPECTRA_PG_DESCRIPTION))

    return np.asarray(r_spectra)


def mp_r_spectra_sector(foto, nb_processes):
    """ Parallel processing of r-spectra

    """
    nb_sample = foto.nb_sampled_frequencies

    # Get radius and sectors within window
    y, x = np.indices((foto.window_size, foto.window_size))
    radius = get_radius(x, y)
    sectors = get_sectors(x, y, foto.nb_sectors, foto.start_sector)

    # Parallel computation of r-spectra
    with mp.Pool(processes=nb_processes) as pool:
        r_spectra = list(tqdm(pool.imap(partial(rspectrum_per_sector,
                                                radius=radius,
                                                sectors=sectors,
                                                window_size=foto.window_size,
                                                nb_sample=nb_sample,
                                                nb_sectors=foto.nb_sectors,
                                                normalized=foto.normalized,
                                                keep_dc_component=foto.keep_dc_component,
                                                no_data_value=foto.gdal_no_data_value),
                                        foto.get_window_generator(), chunksize=MP_CHUNK_SIZE),
                              total=foto.nb_windows, unit_scale=True,
                              desc=R_SPECTRA_SECTOR_PG_DESCRIPTION))
        r_spectra = np.transpose(np.asarray(r_spectra), (1, 0, 2))

    return r_spectra


def h5_incremental_pca(foto, at_random, batch_size, max_iter):
    """ Incremental PCA based on hdf5 dataset chunks

    """
    nb_chunks = foto.nb_windows // foto.chunk_size + min(1, foto.nb_windows % foto.chunk_size)
    population_mean = foto.h5[R_SPECTRA_H5_DS_NAME].attrs["mean"]
    population_std = foto.h5[R_SPECTRA_H5_DS_NAME].attrs["std"]

    if at_random:
        if batch_size is None:
            batch_size = foto.nb_sampled_frequencies * 2**3
        r_spectra_generator = foto.h5.read_at_random(R_SPECTRA_H5_DS_NAME, batch_size)
        nb_iterations = max_iter
    else:
        r_spectra_generator = foto.h5.read(R_SPECTRA_H5_DS_NAME, foto.chunk_size)
        nb_iterations = nb_chunks

    # Incremental PCA
    # Read R-spectra by chunk or batch (if at random) and partial fit
    chunks = tqdm(r_spectra_generator, total=nb_iterations, desc=INC_PCA_PG_DESCRIPTION)
    ipca = incremental_pca(chunks, foto.nb_pca_components, foto.no_data_value, foto.r_spectra_axis,
                           population_mean, population_std, foto.standardized)

    # Finalization
    # Write R-spectra reduced to hdf5 dataset
    foto.h5.reset_dataset(REDUCED_R_SPECTRA_H5_DS_NAME,
                          shape=(foto.nb_windows, foto.nb_pca_components))
    for r_spectra in tqdm(foto.h5.read(R_SPECTRA_H5_DS_NAME, foto.chunk_size),
                          total=nb_chunks, desc=REDUCED_R_SPECTRA_PG_DESCRIPTION):
        r_spectra_reduced = np.full((r_spectra.shape[0], foto.nb_pca_components),
                                    foto.no_data_value, dtype=r_spectra.dtype)
        valid_r_spectra = get_valid_values(r_spectra, foto.no_data_value)
        if foto.standardized:
            valid_r_spectra = (valid_r_spectra - population_mean) / population_std
        try:
            r_spectra_reduced[r_spectra[:, 0] != foto.no_data_value, :] = \
                ipca.transform(valid_r_spectra)
        except ValueError:  # If valid_r_spectra is empty, ipca.transform returns ValueError
            pass
        foto.h5.append(REDUCED_R_SPECTRA_H5_DS_NAME, r_spectra_reduced)

    eigen_vectors = ipca.components_.T
    loadings = eigen_vectors * np.sqrt(ipca.explained_variance_)

    return eigen_vectors, loadings, ipca.explained_variance_ratio_


def h5_incremental_pca_sector(foto, at_random, batch_size, max_iter):
    """ Apply incremental PCA for anisotropic r-spectra based on H5 datasets

    """
    nb_chunks = foto.nb_windows // foto.chunk_size + min(1, foto.nb_windows % foto.chunk_size)
    population_mean = foto.h5[R_SPECTRA_H5_DS_NAME].attrs["mean"]
    population_std = foto.h5[R_SPECTRA_H5_DS_NAME].attrs["std"]

    if at_random:
        if batch_size is None:
            batch_size = foto.nb_sampled_frequencies * 2**3
        r_spectra_generator = foto.h5.read_at_random(R_SPECTRA_H5_DS_NAME,
                                                     batch_size,
                                                     axis=foto.r_spectra_axis)
        nb_iterations = max_iter
    else:
        r_spectra_generator = foto.h5.read(R_SPECTRA_H5_DS_NAME,
                                           foto.chunk_size,
                                           axis=foto.r_spectra_axis)
        nb_iterations = nb_chunks

    # Incremental PCA
    # Read R-spectra by chunk or batch (if at random) and partial fit
    chunks = tqdm(r_spectra_generator, total=nb_iterations, desc=INC_PCA_PG_DESCRIPTION)
    ipca = incremental_pca(chunks, foto.nb_pca_components, foto.no_data_value, foto.r_spectra_axis,
                           population_mean, population_std, foto.standardized, foto.nb_sectors)

    # Finalization
    # Write R-spectra reduced to hdf5 dataset
    foto.h5.reset_dataset(REDUCED_R_SPECTRA_H5_DS_NAME,
                          shape=(foto.nb_sectors, foto.nb_windows, foto.nb_pca_components))
    for r_spectra in tqdm(foto.h5.read(R_SPECTRA_H5_DS_NAME, foto.chunk_size,
                                       axis=foto.r_spectra_axis),
                          total=nb_chunks, desc=REDUCED_R_SPECTRA_PG_DESCRIPTION):
        r_spectra_reduced = np.full((foto.nb_sectors, r_spectra.shape[1], foto.nb_pca_components),
                                    foto.no_data_value, dtype=r_spectra.dtype)
        valid_r_spectra = get_valid_values_3d(r_spectra, foto.no_data_value)
        if foto.standardized:
            valid_r_spectra -= np.expand_dims(population_mean, axis=foto.r_spectra_axis)
            valid_r_spectra /= np.expand_dims(population_std, axis=foto.r_spectra_axis)
        ipca_transform = np.asarray([ipca_.transform(sector_r_spectra)
                                     for ipca_, sector_r_spectra in zip(ipca, valid_r_spectra)])
        r_spectra_reduced[:, r_spectra[0, :, 0] != foto.no_data_value, :] = ipca_transform
        foto.h5.append(REDUCED_R_SPECTRA_H5_DS_NAME,
                       r_spectra_reduced,
                       axis=foto.r_spectra_axis)

    return [ipca_.components_.T for ipca_ in ipca]


def normal_pca(foto):
    """ Normal PCA

    """
    valid_r_spectra = get_valid_values(foto.r_spectra, foto.no_data_value)
    if foto.standardized:
        valid_r_spectra = standardize(valid_r_spectra)

    # Start PCA process
    quit_process = mp.Event()
    p = mp.Process(target=_tqdm_pca, args=(valid_r_spectra.size // 100, quit_process))
    p.start()

    #
    r_spectra_reduced = np.full((foto.nb_windows, foto.nb_pca_components), foto.no_data_value,
                                dtype=foto.r_spectra.dtype)
    eigen_vectors, \
        loadings, \
        r_spectra_reduced[foto.r_spectra[:, 0] != foto.no_data_value, :],\
        explained_variance = \
        pca(valid_r_spectra, foto.nb_pca_components)
    # eigen_vectors, r_spectra_reduced = pca(foto.r_spectra, foto.nb_pca_components)

    # Stop PCA process
    quit_process.set()
    p.join()

    return eigen_vectors, loadings, r_spectra_reduced, explained_variance


def normal_pca_sector(foto):
    """ Compute PCA for each quadrant

    """
    valid_r_spectra = get_valid_values_3d(foto.r_spectra, foto.no_data_value)

    quit_process = mp.Event()
    p = mp.Process(target=_tqdm_pca,
                   args=(foto.nb_sectors * valid_r_spectra.size // 100, quit_process))
    p.start()

    r_spectra_reduced = np.full((foto.nb_sectors, foto.nb_windows, foto.nb_pca_components),
                                foto.no_data_value, dtype=foto.r_spectra.dtype)
    if foto.standardized:
        results = [pca(standardize(rspec), foto.nb_pca_components) for rspec in valid_r_spectra]
    else:
        results = [pca(rspec, foto.nb_pca_components) for rspec in valid_r_spectra]
    eigen_vectors = np.asarray([result[0] for result in results])
    r_spectra_reduced[:, foto.r_spectra[0, :, 0] != foto.no_data_value, :] = \
        np.asarray([result[1] for result in results])

    # Stop PCA process
    quit_process.set()
    p.join()

    return eigen_vectors, r_spectra_reduced


def _tqdm_pca(bar_length, quit_process):
    pg = tqdm(total=bar_length, desc=PCA_PG_DESCRIPTION, unit_scale=True)
    for i in range(int(bar_length)):
        pg.update(1)
        time.sleep(0.0001)
        if quit_process.is_set():
            break
    pg.n = bar_length
    pg.refresh()
    pg.close()



# -*- coding: utf-8 -*-

""" FOTO main module.

Main module of FOTO algorithm. Defines FotoBase class and inherited classes.
"""

import multiprocessing as mp
import os
from abc import abstractmethod

try:
    import gdal
except ModuleNotFoundError:
    from osgeo import gdal

from fototex.base import FotoBase, Batch, Sector
from fototex.plotting import plot_factorial_plan
from fototex.utils import lazyproperty, check_type_in_collection

from fototex.exceptions import FotoBatchError, FotoError, FotoSectorBatchError
from fototex._numba import get_block_windows, get_moving_windows
from fototex.foto_tools import degrees_to_cardinal
from fototex.io import write_rgb


# TODO: add standardize and keep_dc_component options in path names ?


class Foto(FotoBase):
    """ Foto class instance

    Foto object allows to run the Foto algorithm (Couteron et al., 2006) on
    any kind of raster.
    """

    def __init__(self, image, band=None, method="block", in_memory=True,
                 data_chunk_size=50000, *args, **kwargs):
        """ Foto class constructor

        Description
        -----------
        Build a Foto object on which might
        later be run the algorithm. It is
        specifically designed to run on one
        image.

        Parameters
        ----------
        image: str
            path to raster file (must be gdal readable)
        band: int
            band number if multi-band raster
        method: str
            method for window analysis ("block" or "moving")
        in_memory: bool
            if True, import whole raster or band as numpy array
        data_chunk_size: int
            size (nb of elements) of a chunk of data to load in memory (if in_memory == False)
        """
        try:
            self.dataset = gdal.Open(image, gdal.GA_ReadOnly)
        except RuntimeError as e:
            raise FotoError(e)

        if band is None:
            self.band = 1
        else:
            self.band = band

        self.out_dir = os.path.dirname(self.dataset.GetDescription())
        self.method = method
        self.in_memory = in_memory
        self.data_chunk_size = data_chunk_size

    def __del__(self):
        # Explicitly close GDAL dataset
        self.dataset = None

    def _plot_factorial_plan(self, root, window_size, reduced_r_spectra,
                             method, nb_points, data_range, nb_quadrants,
                             norm_method, nb_windows_per_side,
                             main_fig_rel_size, contrast_range, invert_axis,
                             *args, **kwargs):

        plot_factorial_plan(root, self, window_size, reduced_r_spectra, method,
                            nb_points, data_range, nb_quadrants, norm_method,
                            main_fig_rel_size, nb_windows_per_side, contrast_range,
                            invert_axis)

    def get_window_generator(self):
        """ Create window generator depending on the given memory method

        :return: window generator
        """
        if self.in_memory:
            return (self.image[w[1]:w[1] + w[3], w[0]:w[0] + w[2]] for w in self.windows)
        else:
            return (self.dataset.GetRasterBand(self.band).ReadAsArray(*window)
                    for window in self.windows)

    def save_rgb(self):
        """ Save RGB image to file using gdal

        Description
        -----------
        Save reduced r-spectra table to RGB map

        Returns
        -------
        """
        write_rgb(self)

    @property
    def nb_windows(self):
        return self.rgb_height * self.rgb_width

    @lazyproperty
    def image(self):
        if self.in_memory:
            return self.dataset.GetRasterBand(self.band).ReadAsArray()

    @property
    def image_name(self):
        return os.path.splitext(os.path.split(self.dataset.GetDescription())[1])[0]

    @lazyproperty
    def gdal_no_data_value(self):
        return self.dataset.GetRasterBand(self.band).GetNoDataValue()

    @property
    def offset(self):
        return int((self.window_size - 1) / 2)

    @property
    def path(self):
        new_path = super().path
        return new_path + f"image={self.image_name}_"

    @property
    def pixel_size(self):
        return self.dataset.GetGeoTransform()[1]  # TODO: be sure units are meters

    @property
    def rgb_file(self):
        return self.path + "rgb.tif"

    @property
    def rgb_width(self):
        if self.method == "block":
            return int(self.dataset.RasterXSize / self.window_size) #+ \
                   # min(1, self.dataset.RasterXSize % self.window_size)
        else:
            return int((self.dataset.RasterXSize - 2 * self.offset) / self.window_step) #+ \
                   # min(1, self.dataset.RasterXSize % self.window_step)

    @property
    def rgb_height(self):
        if self.method == "block":
            return int(self.dataset.RasterYSize / self.window_size) #+ \
                   # min(1, self.dataset.RasterYSize % self.window_size)
        else:
            return int((self.dataset.RasterYSize - 2 * self.offset) / self.window_step) #+ \
                   # min(1, self.dataset.RasterYSize % self.window_step)

    @property
    def windows(self):
        if self.method == 'block':
            return get_block_windows(self.window_size, self.dataset.RasterXSize,
                                     self.dataset.RasterYSize)
        else:
            return get_moving_windows(self.window_size, self.dataset.RasterXSize,
                                      self.dataset.RasterYSize, self.window_step)


class FotoBatch(Batch):
    """ FotoBatch class

    FotoBatch allows for supplying image batches to the Foto algorithm
    """

    def __init__(self, out_dir, foto_collection, method="block",
                 in_memory=True, data_chunk_size=50000):
        """ Build FotoBatch instance

        Description
        -----------
        FotoBatch instance allows for applying the
        FOTO algorithm on multiple images (batches)

        Parameters
        ----------
        out_dir: str
            path to directory where outputs will be saved
        foto_collection: list or tuple
            collection of Foto instances
        method: str
            sliding window method {'block' or 'moving_window'}
        in_memory: bool
            either implement FOTO in memory or using HDF5 file storage on the fly
        data_chunk_size: int
            if HDF5 storage is implemented, number of data per chunk
        """
        # Store Foto instances of images
        try:
            check_type_in_collection(foto_collection, Foto)
            self.foto_instances = foto_collection
        except TypeError as e:
            raise FotoBatchError(e)

        self.out_dir = out_dir
        self.method = method
        self.in_memory = in_memory
        self.data_chunk_size = data_chunk_size

    def _plot_factorial_plan(self, root, window_size, reduced_r_spectra,
                             method, nb_points, data_range, nb_quadrants,
                             norm_method, nb_windows_per_side,
                             main_fig_rel_size, contrast_range, invert_axis,
                             *args, **kwargs):

        plot_factorial_plan(root, self.foto_instances, window_size, reduced_r_spectra, method,
                            nb_points, data_range, nb_quadrants, norm_method, main_fig_rel_size,
                            nb_windows_per_side, contrast_range, invert_axis)

    def compute_r_spectra(self, window_size, window_step=1, nb_sampled_frequencies=None,
                          normalized=False, keep_dc_component=False, nb_processes=mp.cpu_count(),
                          *args, **kwargs):
        """ Compute r-spectra tables for all batches

        Description
        -----------
        Compute r-spectra table from all supplied
        image batches

        Parameters
        ----------
        window_size: int
            size of window
        window_step: int
            step used in sliding window if method is "moving_window"
        nb_sampled_frequencies: int
            number of frequencies to sample
        normalized: bool
            if True, divide by window variance
        keep_dc_component: bool
            if True, keep DC component in FFT
        nb_processes: int
            number of processes for multiprocessing
        args:
        kwargs:

        Returns
        -------
        FotoBatch:
            the current instance
        """
        for foto in self.foto_instances:
            foto.window_size = window_size
            foto.window_step = window_step
        super().compute_r_spectra(window_size, window_step, nb_sampled_frequencies,
                                  normalized, keep_dc_component, nb_processes)

    @property
    @abstractmethod
    def gdal_no_data_value(self):
        pass


class FotoSector(Foto, Sector):

    def __init__(self, image, nb_sectors=6, start_sector=0, band=None,
                 method="block", in_memory=True, data_chunk_size=50000):
        """ FotoSector constructor

        Description
        -----------
        FotoSector apply the anisotropic version
        of the FOTO algorithm. Typically, depending
        on the required number of sectors, r-spectra
        will be computed for each sector, i.e. circle
        division.

        Parameters
        ----------
        image: str
            valid path to raster image
        nb_sectors: int
            number of sectors
        start_sector: float
            sectors' starting direction (by default: North, i.e 0°)
        band: int
            band number
        method: str
            window sliding method
        in_memory: bool
            if True, store all computations in memory, otherwise use H5 file storage on-the-fly
        data_chunk_size: int
            when using H5 storage, number of data per chunk when reading/writing from/to file

        Note
        ----
        We use here the feature of multiple inheritance,
        by building FotoSector as the mixin of Foto and
        Sector. Look at the Note in Sector to understand
        how to avoid typical issues when doing this. Here,
        Foto and Sector both inherit from FotoBase, as a result
        we must take care of the argument order in the constructor
        of both Foto and Sector classes.
        """
        self.nb_sectors = nb_sectors
        self.start_sector = start_sector

        super().__init__(image, band, method, in_memory, data_chunk_size)

    @property
    def path(self):
        new_path = super().path
        return [new_path + f"sector={sector:.0f}_{degrees_to_cardinal(sector)}_"
                for sector in self.sectors]

    @property
    def rgb_file(self):
        return [path + "rgb.tif" for path in self.path]


class FotoSectorBatch(FotoBatch, Sector):

    def __init__(self, out_dir, foto_collection, nb_sectors=6, start_sector=0,
                 method="block", in_memory=True, data_chunk_size=50000):
        """ FotoSectorBatch

        Description
        -----------
        Run FOTO anisotropic algorithm from
        image batches

        Parameters
        ----------
        out_dir: str
            valid path to output directory where results must be stored
        foto_collection: list or tuple
            collection of FotoSector instances
        nb_sectors: int
            number of sectors
        start_sector: int
            starting sector
        method: str
            valid sliding window method
        in_memory: bool
            if True, all computations are made in memory, otherwise use H5 storage on-the-fly
        data_chunk_size: int
            if H5 storage is used, number of data per chunk when reading/writing from/to file
        """
        super().__init__(out_dir, foto_collection, method, in_memory, data_chunk_size)

        try:
            check_type_in_collection(foto_collection, FotoSector)
        except TypeError as e:
            raise FotoSectorBatchError(e)

    @property
    @abstractmethod
    def gdal_no_data_value(self):
        pass



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

import numpy as np

from itertools import chain, islice

from sklearn.decomposition import PCA

from fototex import R_SPECTRA_NO_DATA_VALUE
from fototex._numba import sector_average, azimuthal_average, zero_padding, only_contains


def azimuth_to_counterclockwise(az, unit="rad"):
    """ Convert azimuth (N=0°, E=90°, S=180°, W=270°)
    to counterclockwise angle (N=90°, E=0°, S=270°, W=180°)

    Parameters
    ----------
    az: float or numpy.ndarray[float]
        Azimuthal angle
    unit: str
        angle unit of azimuth ('rad', 'deg')

    Returns
    -------

    """
    if unit == "rad":
        return (2 * np.pi - (az - np.pi / 2)) % (2 * np.pi)
    else:
        return (360 - (az - 90)) % 360


def counterclockwise_to_azimuth(ccw, unit="rad"):
    """ Convert counterclockwise angle to azimuth

    Parameters
    ----------
    ccw: float or numpy.ndarray
        Counterclockwise angle
    unit: str
        angle unit of counterclockwise angle ('rad', 'deg')

    Returns
    -------

    """
    if unit == "rad":
        return (5 * np.pi/2 - ccw) % (2 * np.pi)
    else:
        return (450 - ccw) % 360


def degrees_to_cardinal(d):
    """ Convert degrees to cardinal direction

    Thanks to https://gist.github.com/RobertSudwarts/acf8df23a16afdb5837f

    Parameters
    ----------
    d: float

    Returns
    -------
    """
    dirs = ["N", "NNE", "NE", "ENE", "E", "ESE", "SE", "SSE",
            "S", "SSW", "SW", "WSW", "W", "WNW", "NW", "NNW"]
    ix = int((d + 11.25)/22.5)
    return dirs[ix % 16]


def get_slice_along_axis(ndim, axis, _slice):
    """ Used to make indexing with any n-dimensional numpy array

    Parameters
    ----------
    ndim: int
        number of dimensions
    axis: int
        axis for which we want the slice
    _slice: slice
        the required slice

    Returns
    -------
    """
    slc = [slice(None)] * ndim
    slc[axis] = _slice

    return tuple(slc)


def get_power_spectrum_density(window, normalized):
    """ Compute power spectrum density for given window

    Description
    -----------

    Parameters
    ----------
    window: numpy.ndarray
        2D window
    normalized: bool
        either divide by window variance or not

    Returns
    -------
    """
    # Fast Fourier Transform (FFT) in 2 dims,
    # center fft and then calculate 2D power spectrum density
    ft = np.fft.fft2(window, norm="ortho")
    ft = np.fft.fftshift(ft)
    psd = np.abs(ft) ** 2

    if normalized:
        var = np.var(window)
        if var != 0:
            psd /= var

    return psd


def pca(data, n_components):
    """ Principal component analysis

    Parameters
    ----------
    data: numpy.ndarray
        normalized data array
    n_components: int
        number of dimensions for PCA

    Returns
    -------
    """

    # replace nodata and inf values and standardize
    # data = np.nan_to_num(data)
    # Standardization (mean=0 and std=1)

    # if sklearn_pca:
    sk_pca = PCA(n_components=n_components)
    sk_pca.fit(data)

    eigen_vectors = sk_pca.components_.T
    loadings = eigen_vectors * np.sqrt(sk_pca.explained_variance_)

    return eigen_vectors, loadings, sk_pca.transform(data), sk_pca.explained_variance_ratio_


def rspectrum(window, radius, window_size, nb_sample, normalized, keep_dc_component, no_data_value):
    """ Compute r-spectrum for given window and filter smaller ones and the ones with no data

    Description
    -----------
    Calculate the azimuthally averaged 1D power
    spectrum (also called radial spectrum, i.e. r-spectrum)

    Parameters
    ----------
    window: numpy.ndarray
        window array
    radius: numpy.ndarray
        corresponding radius integer
    window_size: int
        window typical size
    nb_sample: int
        number of frequencies to sample within window
    normalized: bool
        divide r-spectrum by window variance
    keep_dc_component: bool
        either keep DC component of the FFT (0 frequency part of the signal) or not. It
        may substantially change the results, so use it carefully.
    no_data_value: int or float
        value corresponding to no data

    Returns
    -------
    numpy.ndarray:
        array of isotropic r-spectra
    """
    if only_contains(window, no_data_value):
        return np.full(nb_sample, R_SPECTRA_NO_DATA_VALUE)
    else:
        if no_data_value:
            window = zero_padding(window, no_data_value)
        if keep_dc_component:
            return azimuthal_average(radius,
                                     get_power_spectrum_density(window, normalized))[0:nb_sample]
        else:
            return azimuthal_average(radius,
                                     get_power_spectrum_density(window,
                                                                normalized))[1:nb_sample + 1]


def rspectrum_per_sector(window, radius, sectors, window_size, nb_sample,
                         nb_sectors, normalized, keep_dc_component, no_data_value):
    """ Compute r-spectrum for each sector

    Description
    -----------

    Calculate radial spectrum in specific directions, i.e. quadrants

    Parameters
    ----------

    window: numpy.array
        window
    radius:
    sectors: numpy.array
        result from get_sectors function
        (divide the circle into sectors according to nb_sectors)
    window_size: int
        window typical size
    nb_sample: int
        number of frequencies to sample within window
    nb_sectors: int
        number of sectors
    normalized: bool
        divide r-spectrum by window variance
    keep_dc_component: bool
        either keep DC component of the FFT (0 frequency part of the signal) or not. It
        may substantially change the results, so use it carefully.
    no_data_value: int
        value corresponding to no data

    Returns
    -------
    """
    if only_contains(window, no_data_value):
        return np.full((nb_sectors, nb_sample), R_SPECTRA_NO_DATA_VALUE)
    else:
        if no_data_value:
            window = zero_padding(window, no_data_value)
        if keep_dc_component:
            return sector_average(get_power_spectrum_density(window, normalized),
                                  radius, sectors, nb_sectors)[:, 0:nb_sample]
        else:
            return sector_average(get_power_spectrum_density(window, normalized),
                                  radius, sectors, nb_sectors)[:, 1:nb_sample + 1]


def scale_range(array, lower, upper, no_data=None, axis=0):
    """ Scale array to range values

    Parameters
    ----------
    array
    lower: int or float
        lower bound of range
    upper: int or float
        upper bound of range
    no_data
    axis

    Returns
    -------

    """
    if no_data is not None:
        return (array - np.min(array[array != no_data], axis=axis)) * (upper - lower) / \
               (np.max(array[array != no_data], axis=axis) - np.min(array[array != no_data],
                                                                    axis=axis)) + lower
    else:
        return (array - np.min(array, axis=axis)) * (upper - lower) / \
               (np.max(array, axis=axis) - np.min(array, axis=axis)) + lower


def split_into_chunks(iterable, size=10):
    """ Split iterable into chunks

    Parameters
    ----------
    iterable: iterable
    size: int
        size of each chunk

    Returns
    -------
    """
    iterator = iter(iterable)
    for first in iterator:
        yield chain([first], islice(iterator, size - 1))


def standard_deviation(nb_data, sum_of_values, sum_of_square_values):
    """ Compute standard deviation based on variance formula

    Parameters
    ----------
    nb_data: int
        number of data
    sum_of_values:
    sum_of_square_values:

    Returns
    -------
    """
    return np.sqrt(sum_of_square_values / nb_data - (sum_of_values / nb_data) ** 2)


def standardize(data):
    """ Standardize data (subtract mean and divide by std)

    """
    return (data - data.mean(axis=0)) / data.std(axis=0)


def get_window(idx, window_size, raster_x_size,
               raster_y_size, method, step=1):
    """ Get window corresponding to position index

    Parameters
    ----------
    idx: int
        position index
    window_size: int
        Window size
    raster_x_size: int
        Raster X size
    raster_y_size: int
        Raster Y size
    method: str
        Sliding window method {'block', 'moving'}
    step: int, default 1
        Step used in moving window method

    Returns
    -------
    tuple

    """
    if method == "block":
        n_x = raster_x_size // window_size  # + min(1, raster_x_size % window_size)
        n_y = raster_y_size // window_size  # + min(1, raster_y_size % window_size)
        if idx >= (n_x * n_y):
            raise ValueError(f"Position index is out of bounds: {idx} >= {n_x*n_y}")
        y1 = window_size * min(idx // n_x, n_y - 1)
        x1 = window_size * min(idx - (idx // n_x) * n_x, n_x - 1)
        x_size = min(window_size, raster_x_size - x1)
        y_size = min(window_size, raster_y_size - y1)
    else:
        offset = int((window_size - 1) / 2)
        n_x = raster_x_size // step + min(1, raster_x_size % step)
        n_y = raster_y_size // step + min(1, raster_y_size % step)
        if idx >= (n_x * n_y):
            raise ValueError(f"Position index is out of bounds: {idx} >= {n_x*n_y}")
        y = step * (idx // n_x)
        x = (idx - (idx // n_x) * n_x) * step
        x1 = max(0, x - offset)
        y1 = max(0, y - offset)
        x2 = min(raster_x_size - 1, x + offset)
        y2 = min(raster_y_size - 1, y + offset)
        x_size = (x2 - x1) + 1
        y_size = (y2 - y1) + 1

    return int(x1), int(y1), int(x_size), int(y_size)


def get_window_over_multiple_images(idx, window_size, raster_x_size,
                                    raster_y_size, method, step=1):
    """ Get window corresponding to global position index
     over multiple images

    Parameters
    ----------
    idx
    window_size
    raster_x_size
    raster_y_size
    method
    step

    Returns
    -------

    """

    coords = None
    raster_id = 0

    for x_size, y_size in zip(raster_x_size, raster_y_size):
        try:
            coords = get_window(idx, window_size, x_size, y_size, method, step)
            break
        except ValueError:
            if method == "block":
                n_x = x_size // window_size  # + min(1, x_size % window_size)
                n_y = y_size // window_size  # + min(1, y_size % window_size)
            else:
                n_x = x_size // step + min(1, x_size % step)
                n_y = y_size // step + min(1, y_size % step)

            idx -= int(n_x * n_y)
            raster_id += 1

    return raster_id, coords



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
import os
# from collections import Collection


def check_string(string, list_of_strings):
    """ Check validity of and return string against list of valid strings

    :param string: searched string
    :param list_of_strings: list/tuple/set of valid strings string is to be checked against
    :return: validate string from list of strings if match
    """
    # check_type(string, str, list_of_strings, Collection)
    [check_type(x, str) for x in list_of_strings]

    output_string = []

    for item in list_of_strings:
        if item.lower().startswith(string.lower()):
            output_string.append(item)

    if len(output_string) == 1:
        return output_string[0]
    elif len(output_string) == 0:
        raise ValueError("input must match one of those: {}".format(list_of_strings))
    elif len(output_string) > 1:
        raise ValueError("input match more than one valid value among {}".format(list_of_strings))


def check_type(*args):
    """Check type of arguments

    :param args: tuple list of argument/type
    :return:
    """
    if len(args) % 2 == 0:
        for item in range(0, len(args), 2):
            if not isinstance(args[item], args[item + 1]):
                raise TypeError("Type of argument {} is '{}' but must be '{}'".format(
                    item//2 + 1, type(args[item]).__name__, args[item + 1].__name__))


def check_type_in_collection(collection, _type, include=True):
    """ Check type of elements in collection

    Check type of elements in collection through including or
    excluding specific types.

    :param collection: list/tuple/set/ndarray/etc. collection of data
    :param _type: type or tuple list of types
    :param include: should collection include or exclude type ?
    :return: none
    """
    # check_type(collection, Collection, _type, (type, tuple), include, bool)
    if include is True and not all(isinstance(element, _type) for element in collection):
        raise TypeError("All elements in collection '{}' "
                        "must be {}".format(type(collection), _type))
    elif include is False and any(isinstance(element, _type) for element in collection):
        raise TypeError("Collection '{}' cannot contain {}".format(type(collection), _type))


def isdir(path):
    try:
        return os.path.isdir(path)
    except (TypeError, ValueError):
        return False


def lazyproperty(func):
    name = '_lazy_' + func.__name__

    @property
    def lazy(self):
        if hasattr(self, name):
            return getattr(self, name)
        else:
            value = func(self)
            setattr(self, name, value)
            return value
    return lazy



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

import os
import tempfile

import h5py
import numpy as np
from tqdm import tqdm

from fototex import GDAL_DRIVER, GDAL_FLOAT32, WRITE_RGB_PG_DESCRIPTION
from fototex.exceptions import H5FileAppendError
from fototex.foto_tools import get_slice_along_axis


class H5File:

    def __init__(self, path):
        self.h5 = h5py.File(path, 'a')
        self.path = self.h5.filename
        # Personal note: never put the following line in the class
        # definition (i.e. defined as a class attribute) !! As the
        # dictionary is a mutable object, a modification in any
        # instance will override in all the other instances !!
        self.row_caret_pos = {}

    def __del__(self):
        self.close()

    def __getitem__(self, item):
        if self.has_dataset(item):
            return self.h5[item]

    def append(self, dataset_name, data, axis=0):
        """ Append data to dataset (by row)

        :param dataset_name:
        :param data: data to append to dataset
        :param axis: dataset axis on which append data
        :return:
        """
        if self.has_dataset(dataset_name):
            if dataset_name not in self.row_caret_pos.keys():
                self.row_caret_pos[dataset_name] = 0
            start = self.row_caret_pos[dataset_name]
            self.row_caret_pos[dataset_name] += data.shape[axis]
            idx = get_slice_along_axis(data.ndim, axis,
                                       slice(start, self.row_caret_pos[dataset_name]))
            try:
                self.h5[dataset_name][idx] = data
            except IndexError as e:
                raise H5FileAppendError(e)
            # self.h5[dataset_name][start:self.row_caret_pos[dataset_name], :] = data

    def close(self):
        self.h5.close()

    def copy(self, h5, dataset_name):
        """ Copy H5File dataset to current file

        Dataset must exist in other H5File and must not exist in current file
        :param h5: H5File
        :param dataset_name: name of the corresponding dataset
        :return:
        """
        if h5.has_dataset(dataset_name) and not self.has_dataset(dataset_name):
            self.h5.create_dataset(dataset_name,
                                   data=h5[dataset_name],
                                   shape=h5[dataset_name].shape,
                                   dtype=h5[dataset_name].dtype)

    def create_dataset(self, name, shape, dtype=np.float32):
        """ Create dataset

        :param name:
        :param shape:
        :param dtype:
        :return:
        """
        if not self.has_dataset(name):
            self.h5.create_dataset(name, shape=shape, dtype=dtype)

    def has_dataset(self, name):
        """ Test if h5 file has dataset

        :param name:
        :return:
        """
        return name in self.dataset_names

    def read(self, dataset_name, chunk_size, axis=0):
        """ Sequentially read dataset by row with respect to batch size

        Last batch shall have size <= batch_size

        Parameters
        ----------
        dataset_name: str
        chunk_size: int
        axis: int

        Returns
        -------
        """
        nb_iter = self.h5[dataset_name].shape[axis] // chunk_size + \
            min(1, self.h5[dataset_name].shape[axis] % chunk_size)
        if self.has_dataset(dataset_name):
            for i in range(0, nb_iter):
                yield self.h5[dataset_name][get_slice_along_axis(self.h5[dataset_name].ndim, axis,
                                                                 slice(i * chunk_size,
                                                                       (i+1) * chunk_size))]
                # yield self.h5[dataset_name][i * chunk_size:(i + 1) * chunk_size, :]

    def read_at(self, dataset_name, indices, axis=0):
        """ Read data in dataset at given indices

        Parameters
        ----------
        dataset_name: str
            Name of the dataset
        indices: numpy.ndarray
            indices for which must be returned data in dataset
        axis: int
            Axis for which indices are defined

        Returns
        -------

        """
        if self.has_dataset(dataset_name):
            return self.h5[dataset_name][get_slice_along_axis(self.h5[dataset_name].ndim,
                                                              axis, indices)]

    def read_at_random(self, dataset_name, batch_size, axis=0, max_iter=1000):
        """ Read dataset at random row by row with respect to batch size

        Parameters
        ----------
        dataset_name: str
            Dataset name
        batch_size: int
            Size of each batch to read from dataset
        axis: int
            Axis along which data must be read row by row
        max_iter: int
            Maximum number of iterations for batch process

        Returns
        -------
        """
        rng = np.random.default_rng()
        if self.has_dataset(dataset_name):
            for _ in range(max_iter):
                items = sorted(rng.choice(self.h5[dataset_name].shape[axis],
                                          size=batch_size,
                                          replace=False,
                                          shuffle=False))
                yield self.h5[dataset_name][get_slice_along_axis(self.h5[dataset_name].ndim,
                                                                 axis, items)]
                # yield self.h5[dataset_name][items, :]

    def remove_dataset(self, dataset_name):
        """ Remove dataset within h5 file

        Parameters
        ----------
        dataset_name:

        Returns
        -------
        """
        if self.has_dataset(dataset_name):
            self.reset_caret(dataset_name)
            del self.h5[dataset_name]

    def reset_caret(self, dataset_name):
        try:
            self.row_caret_pos.pop(dataset_name)
        except KeyError:
            pass

    def reset_dataset(self, dataset_name, shape, dtype=np.float32):
        """ Reset dataset

        Parameters
        ----------
        dataset_name: str
            Dataset name
        shape: tuple
            Shape of the dataset table
        dtype:

        Returns
        -------
        """
        self.remove_dataset(dataset_name)
        self.create_dataset(dataset_name, shape, dtype)

    @property
    def dataset_names(self):
        return list(self.h5.keys())

    @property
    def attributes(self):
        return self.h5.attrs

    @property
    def attribute_names(self):
        return list(self.h5.attrs.keys())


class H5TempFile(H5File):

    def __init__(self):

        super().__init__(tempfile.mkstemp(suffix='.h5')[1])

    def __del__(self):
        super().__del__()
        try:
            os.remove(self.path)
        except FileNotFoundError:
            pass


def write_h5(foto, data, ds_name, attr_names=None, attr_values=None):
    """ Write FOTO data into H5 file

    Parameters
    ----------
    foto
    data
    ds_name
    attr_names
    attr_values

    Returns
    -------

    """
    foto.h5_file.remove_dataset(ds_name)
    if foto.in_memory:
        foto.h5_file.create_dataset(ds_name,
                                    shape=data.shape,
                                    dtype=data.dtype)
        foto.h5_file.append(ds_name,
                            data)
    else:
        foto.h5_file.copy(foto.h5, ds_name)

    if attr_names is not None:
        for name, value in zip(attr_names, attr_values):
            foto.h5_file[ds_name].attrs[name] = value


def write_h5_scalar(foto, value, ds_name):
    """ Write FOTO scalar value into H5 file

    Parameters
    ----------
    foto
    value
    ds_name

    Returns
    -------

    """

    foto.h5_file.remove_dataset(ds_name)
    foto.h5_file[ds_name] = value


def write_rgb(foto):
    """ Write RGB image to raster file using gdal

    Write image to raster line by line
    """

    def _write_rgb(rgb_file, r_spectra_reduced):

        try:
            os.remove(rgb_file)
        except FileNotFoundError:
            pass

        # Open or create output dataset
        out_dataset = GDAL_DRIVER.Create(rgb_file,
                                         foto.rgb_width,
                                         foto.rgb_height,
                                         foto.nb_pca_components,
                                         GDAL_FLOAT32)

        topleftx, pxsizex, rotx, toplefty, roty, pxsizey = foto.dataset.GetGeoTransform()

        if foto.method == "block":
            out_dataset.SetGeoTransform((topleftx, pxsizex * foto.window_size, rotx, toplefty, roty,
                                         pxsizey * foto.window_size))
        else:
            out_dataset.SetGeoTransform((topleftx + foto.offset * pxsizex,
                                         pxsizex * foto.window_step, rotx,
                                         toplefty + foto.offset * pxsizey,
                                         roty, pxsizey * foto.window_step))
            # out_dataset.SetGeoTransform(foto.dataset.GetGeoTransform())
        out_dataset.SetProjection(foto.dataset.GetProjection())

        # Set no data value
        for band in range(foto.nb_pca_components):
            out_dataset.GetRasterBand(band + 1).SetNoDataValue(foto.no_data_value)

        # Write to dataset
        row_width = out_dataset.RasterXSize
        for y in tqdm(range(0, out_dataset.RasterYSize),
                      desc=WRITE_RGB_PG_DESCRIPTION % out_dataset.GetDescription(),
                      unit_scale=True):
            for band in range(foto.nb_pca_components):
                rgb = np.expand_dims(r_spectra_reduced[y * row_width:(y + 1) * row_width, band],
                                     axis=0)
                out_dataset.GetRasterBand(band + 1).WriteArray(rgb, 0, y)

        # Close dataset
        out_dataset = None

    try:
        _write_rgb(foto.rgb_file, foto.r_spectra_reduced)
    except (RuntimeError, TypeError):
        for file, r_spectra in zip(foto.rgb_file, foto.r_spectra_reduced):
            _write_rgb(file, r_spectra)



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""


class FotoBaseError(Exception):
    pass


class FotoError(FotoBaseError):
    pass


class FotoBatchError(FotoError):
    pass


class FotoSectorBatchError(FotoBaseError):
    pass


class H5FileError(Exception):
    pass


class H5FileAppendError(H5FileError):
    pass


class ImportSklearnWarning(Warning):
    pass



# -*- coding: utf-8 -*-

""" Compute Fourier Transform Textural Ordination.

FOTO algorithm package by and for python. Use to retrieve
textural information in satellite images.

- Example of usage:
>>> from fototex.foto import Foto
>>> foto = Foto("/path/to/image.tif", method="moving_window")
>>> foto.run(window_size=11)

"""

__version__ = '1.5.9'
__licence__ = "MIT"

try:
    import gdal
except ModuleNotFoundError:
    from osgeo import gdal

# Raise Python exceptions for gdal errors
gdal.UseExceptions()

# Multiprocessing variables
MP_CHUNK_SIZE = 5000

# R-spectra table axis (depending on number of dimensions)
# Isotropic: 2 dimensions, anisotropic: 3 dimensions
# Only advised users might modify it
_ISOTROPIC_R_SPECTRA_AXIS = 0
_ANISOTROPIC_R_SPECTRA_AXIS = 1
_ISOTROPIC_NB_SAMPLE_AXIS = 1
_ANISOTROPIC_NB_SAMPLE_AXIS = 2

# GDAL drivers and data type
GDAL_DRIVER = gdal.GetDriverByName("GTiff")
GDAL_FLOAT32 = gdal.GetDataTypeByName('Float32')

# Main variables
NB_PCA_COMPONENTS = 3
MAX_NB_OF_SAMPLED_FREQUENCIES = 29
MAX_NB_SECTORS = 16
R_SPECTRA_NO_DATA_VALUE = -999

# Progress descriptions
PCA_TRANSFORM_PG_DESCRIPTION = "Reduce R-spectra"
PCA_PG_DESCRIPTION = "Run Principal Component Analysis"
INC_PCA_PG_DESCRIPTION = "Run incremental PCA"
R_SPECTRA_PG_DESCRIPTION = "Retrieve isotropic R-spectra"
R_SPECTRA_SECTOR_PG_DESCRIPTION = "Retrieve anisotropic R-spectra"
REDUCED_R_SPECTRA_PG_DESCRIPTION = "Save reduced R-spectra"
WRITE_RGB_PG_DESCRIPTION = "Write RGB output image to '%s'"

# HDF5 variables
EIGEN_VECTOR_H5_DS_NAME = "eigen_vectors"
EXPLAINED_VARIANCE_RATIO_DS_NAME = "explained_variance_ratio"
LOADINGS_H5_DS_NAME = "loadings"
R_SPECTRA_H5_DS_NAME = "r-spectra"
REDUCED_R_SPECTRA_H5_DS_NAME = "r-spectra-reduced"
WINDOW_SIZE_ATTR_NAME = "window-size"
METHOD_ATTR_NAME = "method"



from setuptools import setup, find_packages

import fototex

with open("README.md", 'r') as fh:
    long_description = fh.read()

with open("requirements.txt") as req:
    install_req = req.read().splitlines()

setup(name='fototex',
      version=fototex.__version__,
      description='Fourier Transform Textural Ordination',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://framagit.org/benjaminpillot/fototex',
      author='Benjamin Pillot <benjamin.pillot@ird.fr>, '
             'Dominique Lyszczarz <observatoire@causses-et-cevennes.fr>, '
             'Claire Teillet, <teillet.claire@hotmail.com>, '
             'Pierre Couteron <pierre.couteron@ird.fr>, '
             'Nicolas Barbier <nicolas.barbier@ird.fr>, '
             'Philippe Verley <philippe.verley@ird.fr>, '
             'Marc Lang <marc.lang@irstea.fr>, '
             'Thibault Catry <thibault.catry@ird.fr>, '
             'Laurent Demagistri <laurent.demagistri@ird.fr>, '
             'Nadine Dessay <nadine.dessay@ird.fr>',
      install_requires=install_req,
      python_requires='>=3',
      license='The MIT Licence',
      packages=find_packages(exclude=("tutorial",)),
      zip_safe=False)
